# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
from Path.Tool.library import Library
from Path.Tool.toolbit import ToolBitEndmill, ToolBitDrill, ToolBitVBit
from Path.Tool.shape import (
    ToolBitShapeEndmill,
    ToolBitShapeDrill,
    ToolBitShapeVBit,
)


class TestPathToolLibrary(unittest.TestCase):
    def test_init(self):
        library = Library("Test Library")
        self.assertEqual(library.label, "Test Library")
        self.assertIsNotNone(library.id)
        self.assertEqual(len(library._bits), 0)
        self.assertEqual(len(library._bit_nos), 0)

    def test_str(self):
        library = Library("My Library", "123-abc")
        self.assertEqual(str(library), '123-abc "My Library"')

    def test_eq(self):
        library1 = Library("Lib1", "same-id")
        library2 = Library("Lib2", "same-id")
        library3 = Library("Lib3", "different-id")
        self.assertEqual(library1, library2)
        self.assertNotEqual(library1, library3)

    def test_iter(self):
        library = Library("Test Library")
        # Create ToolBitShape instances with required parameters
        shape1 = ToolBitShapeEndmill(
            id="dummy_endmill",
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        shape2 = ToolBitShapeDrill(
            id="dummy_drill",
            Diameter=5.0,
            Length=40.0,
            ShankDiameter=5.0,
            Flutes=2,
            TipAngle=118.0,
        )
        bit1 = ToolBitEndmill(shape1)
        bit2 = ToolBitDrill(shape2)
        library.add_bit(bit1)
        library.add_bit(bit2)
        _bits_list = list(library)
        self.assertEqual(len(_bits_list), 2)
        self.assertIn(bit1, _bits_list)
        self.assertIn(bit2, _bits_list)

    def test_get_next_bit_no(self):
        library = Library("Test Library")
        self.assertEqual(library.get_next_bit_no(), 1)
        # Using ToolBit instances in _bit_nos with ToolBitShape
        shape_a = ToolBitShapeEndmill(
            id="dummy_a",
            CuttingEdgeHeight=1.0,
            Diameter=1.0,
            Flutes=1,
            Length=10.0,
            ShankDiameter=1.0,
        )
        shape_b = ToolBitShapeDrill(
            id="dummy_b",
            Diameter=2.0,
            Length=20.0,
            ShankDiameter=2.0,
            Flutes=2,
            TipAngle=118.0,
        )
        shape_c = ToolBitShapeVBit(
            id="dummy_c",
            Diameter=3.0,
            Angle=90.0,
            Length=30.0,
            ShankDiameter=3.0,
            CuttingEdgeAngle=30.0,
            CuttingEdgeHeight=10.0,
            Flutes=2,
            TipDiameter=1.0,
        )
        library._bit_nos = {
            1: ToolBitEndmill(shape_a),
            5: ToolBitDrill(shape_b),
            2: ToolBitVBit(shape_c),
        }
        self.assertEqual(library.get_next_bit_no(), 6)
        library._bit_nos = {}
        self.assertEqual(library.get_next_bit_no(), 1)

    def test_get_bit_no_from_bit(self):
        library = Library("Test Library")
        shape1 = ToolBitShapeEndmill(
            id="dummy_endmill_1",
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        shape2 = ToolBitShapeDrill(
            id="dummy_drill_1",
            Diameter=5.0,
            Length=40.0,
            ShankDiameter=5.0,
            Flutes=2,
            TipAngle=118.0,
        )
        bit1 = ToolBitEndmill(shape1)
        bit2 = ToolBitDrill(shape2)
        library.add_bit(bit1, 1)
        library.add_bit(bit2, 2)
        self.assertEqual(library.get_bit_no_from_bit(bit1), 1)
        self.assertEqual(library.get_bit_no_from_bit(bit2), 2)
        shape_cutter = ToolBitShapeVBit(
            id="dummy_cutter_1",
            Diameter=3.0,
            Angle=90.0,
            Length=30.0,
            ShankDiameter=3.0,
            CuttingEdgeAngle=30.0,
            CuttingEdgeHeight=10.0,
            Flutes=2,
            TipDiameter=1.0,
        )
        self.assertIsNone(library.get_bit_no_from_bit(ToolBitVBit(shape_cutter)))

    def test_assign_new_bit_no(self):
        library = Library("Test Library")
        shape1 = ToolBitShapeEndmill(
            id="dummy_endmill_2",
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        shape2 = ToolBitShapeDrill(
            id="dummy_drill_2",
            Diameter=5.0,
            Length=40.0,
            ShankDiameter=5.0,
            Flutes=2,
            TipAngle=118.0,
        )
        bit1 = ToolBitEndmill(shape1)
        bit2 = ToolBitDrill(shape2)

        # Assign bit1 without specifying number (should get 1)
        library.add_bit(bit1)
        self.assertEqual(len(library._bits), 1)
        self.assertEqual(len(library._bit_nos), 1)
        self.assertIn(1, library._bit_nos)
        self.assertEqual(library._bit_nos[1], bit1)

        # Assign bit2 to number 1 (should reassign bit1)
        library.add_bit(bit2, 1)
        self.assertEqual(len(library._bits), 2)
        self.assertEqual(len(library._bit_nos), 2)
        self.assertIn(1, library._bit_nos)
        self.assertEqual(library._bit_nos[1], bit2)
        # Check if bit1 was reassigned to a new bit number (should be 2)
        self.assertIn(2, library._bit_nos)
        self.assertEqual(library._bit_nos[2], bit1)

        # Assign bit2 to number 10
        library.assign_new_bit_no(bit2, 10)
        self.assertEqual(len(library._bits), 2)
        self.assertEqual(len(library._bit_nos), 2)
        self.assertIn(10, library._bit_nos)
        self.assertEqual(library._bit_nos[10], bit2)
        self.assertNotIn(1, library._bit_nos)  # bit2 should no longer be at 1
        self.assertIn(2, library._bit_nos)
        self.assertEqual(library._bit_nos[2], bit1)

        # Assign bit1 to number 5
        library.assign_new_bit_no(bit1, 5)
        self.assertEqual(len(library._bits), 2)
        self.assertEqual(len(library._bit_nos), 2)
        self.assertIn(5, library._bit_nos)
        self.assertEqual(library._bit_nos[5], bit1)
        self.assertNotIn(2, library._bit_nos)  # bit1 should no longer be at 2
        self.assertIn(10, library._bit_nos)
        self.assertEqual(library._bit_nos[10], bit2)

    def test_add_bit(self):
        library = Library("Test Library")
        shape1 = ToolBitShapeEndmill(
            id="dummy_endmill_3",
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        shape2 = ToolBitShapeDrill(
            id="dummy_drill_3",
            Diameter=5.0,
            Length=40.0,
            ShankDiameter=5.0,
            Flutes=2,
            TipAngle=118.0,
        )
        bit1 = ToolBitEndmill(shape1)
        bit2 = ToolBitDrill(shape2)

        library.add_bit(bit1)
        self.assertEqual(len(library._bits), 1)
        self.assertIn(bit1, library._bits)
        self.assertEqual(len(library._bit_nos), 1)
        self.assertIn(1, library._bit_nos)
        self.assertEqual(library._bit_nos[1], bit1)

        library.add_bit(bit2, 5)
        self.assertEqual(len(library._bits), 2)
        self.assertIn(bit1, library._bits)
        self.assertIn(bit2, library._bits)
        self.assertEqual(len(library._bit_nos), 2)
        self.assertIn(1, library._bit_nos)
        self.assertEqual(library._bit_nos[1], bit1)
        self.assertIn(5, library._bit_nos)
        self.assertEqual(library._bit_nos[5], bit2)

        # Add bit1 again (should not increase bit count in _bits list)
        library.add_bit(bit1, 10)
        self.assertEqual(len(library._bits), 2)
        self.assertIn(bit1, library._bits)
        self.assertIn(bit2, library._bits)
        self.assertEqual(len(library._bit_nos), 2)  # _bit_nos count remains 2
        self.assertIn(10, library._bit_nos)
        self.assertEqual(library._bit_nos[10], bit1)
        self.assertIn(5, library._bit_nos)
        self.assertEqual(library._bit_nos[5], bit2)
        self.assertNotIn(1, library._bit_nos)  # bit1 should no longer be at 1

    def test_get_bits(self):
        library = Library("Test Library")
        shape1 = ToolBitShapeEndmill(
            id="dummy_endmill_4",
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        shape2 = ToolBitShapeDrill(
            id="dummy_drill_4",
            Diameter=5.0,
            Length=40.0,
            ShankDiameter=5.0,
            Flutes=2,
            TipAngle=118.0,
        )
        bit1 = ToolBitEndmill(shape1)
        bit2 = ToolBitDrill(shape2)
        self.assertEqual(library.get_bits(), [])
        library.add_bit(bit1)
        library.add_bit(bit2)
        _bits_list = library.get_bits()
        self.assertEqual(len(_bits_list), 2)
        self.assertIn(bit1, _bits_list)
        self.assertIn(bit2, _bits_list)

    def test_has_bit(self):
        library = Library("Test Library")
        shape1 = ToolBitShapeEndmill(
            id="dummy_endmill_5",
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        shape2 = ToolBitShapeDrill(
            id="dummy_drill_5",
            Diameter=5.0,
            Length=40.0,
            ShankDiameter=5.0,
            Flutes=2,
            TipAngle=118.0,
        )
        bit1 = ToolBitEndmill(shape1)
        bit2 = ToolBitDrill(shape2)
        library.add_bit(bit1)
        self.assertTrue(library.has_bit(bit1))
        self.assertFalse(library.has_bit(bit2))
        # Create a new ToolBit with the same properties but different instance
        shape1_copy = ToolBitShapeEndmill(
            id="dummy_endmill_5_copy",  # Use a different ID for the copy
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        bit1_copy = ToolBitEndmill(shape1_copy)
        self.assertFalse(library.has_bit(bit1_copy))

    def test_remove_bit(self):
        library = Library("Test Library")
        shape1 = ToolBitShapeEndmill(
            id="dummy_endmill_6",
            CuttingEdgeHeight=10.0,
            Diameter=10.0,
            Flutes=2,
            Length=50.0,
            ShankDiameter=10.0,
        )
        shape2 = ToolBitShapeDrill(
            id="dummy_drill_6",
            Diameter=5.0,
            Length=40.0,
            ShankDiameter=5.0,
            Flutes=2,
            TipAngle=118.0,
        )
        shape3 = ToolBitShapeVBit(
            id="dummy_cutter_6",
            Diameter=3.0,
            Angle=90.0,
            Length=30.0,
            ShankDiameter=3.0,
            CuttingEdgeAngle=30.0,
            CuttingEdgeHeight=10.0,
            Flutes=2,
            TipDiameter=1.0,
        )
        bit1 = ToolBitEndmill(shape1)
        bit2 = ToolBitDrill(shape2)
        bit3 = ToolBitVBit(shape3)

        library.add_bit(bit1, 1)
        library.add_bit(bit2, 2)
        library.add_bit(bit3, 3)
        self.assertEqual(len(library._bits), 3)
        self.assertEqual(len(library._bit_nos), 3)

        library.remove_bit(bit2)
        self.assertEqual(len(library._bits), 2)
        self.assertNotIn(bit2, library._bits)
        self.assertEqual(len(library._bit_nos), 2)
        self.assertNotIn(2, library._bit_nos)
        self.assertNotIn(bit2, library._bit_nos.values())

        library.remove_bit(bit1)
        self.assertEqual(len(library._bits), 1)
        self.assertNotIn(bit1, library._bits)
        self.assertEqual(len(library._bit_nos), 1)
        self.assertNotIn(1, library._bit_nos)
        self.assertNotIn(bit1, library._bit_nos.values())

        library.remove_bit(bit3)
        self.assertEqual(len(library._bits), 0)
        self.assertNotIn(bit3, library._bits)
        self.assertEqual(len(library._bit_nos), 0)
        self.assertNotIn(3, library._bit_nos)
        self.assertNotIn(bit3, library._bit_nos.values())

        # Removing a non-existent bit should not raise an error
        shape_nonexistent = ToolBitShapeEndmill(
            id="dummy_nonexistent_6",
            CuttingEdgeHeight=99.0,
            Diameter=99.0,
            Flutes=1,
            Length=99.0,
            ShankDiameter=99.0,
        )
        library.remove_bit(ToolBitEndmill(shape_nonexistent))
        self.assertEqual(len(library._bits), 0)
        self.assertEqual(len(library._bit_nos), 0)


if __name__ == "__main__":
    unittest.main()



"""
LazyLoader will defer import of a module until first usage. Usage:
from lazy_loader.lazy_loader import LazyLoader
numpy = LazyLoader("numpy", globals(), "numpy")

or

whatever = LazyLoader("module", globals(), "module.whatever")

or to replicate import module as something

something = LazyLoader("module", globals(), "module")
"""


# SPDX-License-Identifier: LGPL-2.1-or-later

# Unit tests for the Path.Tool.Shape module and its document utilities.
import unittest
from unittest.mock import patch, MagicMock, call
from Path.Tool.shape import doc
import os


mock_freecad = MagicMock(Name="FreeCAD_Mock")
mock_freecad.Console = MagicMock()
mock_freecad.Console.PrintWarning = MagicMock()
mock_freecad.Console.PrintError = MagicMock()

mock_obj = MagicMock(Name="Object_Mock")
mock_obj.Label = "MockObjectLabel"
mock_obj.Name = "MockObjectName"

mock_doc = MagicMock(Name="Document_Mock")
mock_doc.Objects = [mock_obj]


class TestPathToolShapeDoc(unittest.TestCase):
    def setUp(self):
        """Reset mocks before each test."""
        # Resetting the top-level mock recursively resets its children
        # (newDocument, getDocument, openDocument, closeDocument, Console, etc.)
        # and their call counts, return_values, side_effects.
        mock_freecad.reset_mock()
        mock_doc.reset_mock()
        mock_obj.reset_mock()

        # Re-establish default state/attributes potentially cleared by reset_mock
        # or needed for tests.
        mock_doc.Objects = [mock_obj]
        mock_obj.Label = "MockObjectLabel"
        mock_obj.Name = "MockObjectName"
        mock_obj.getTypeIdOfProperty = MagicMock(return_value="App::PropertyString")
        # Ensure mock_doc also has a Name attribute used in tests/code
        mock_doc.Name = "Document_Mock"  # Used in closeDocument calls

        # Clear attributes potentially added by setattr in previous tests.
        # reset_mock() doesn't remove attributes added this way.
        # Focus on attributes known to be added by tests in this file.
        for attr_name in ["Diameter", "Length", "Height"]:
            if hasattr(mock_obj, attr_name):
                try:
                    delattr(mock_obj, attr_name)
                except AttributeError:
                    pass  # Ignore if already gone

    """Tests for the document utility functions in Path/Tool/Shape/doc.py"""

    def test_doc_find_shape_object_body_priority(self):
        """Test find_shape_object prioritizes PartDesign::Body."""
        body_obj = MagicMock(Name="Body_Mock")
        body_obj.isDerivedFrom = lambda typeName: typeName == "PartDesign::Body"
        part_obj = MagicMock(Name="Part_Mock")
        part_obj.isDerivedFrom = lambda typeName: typeName == "Part::Feature"
        mock_doc.Objects = [part_obj, body_obj]
        found = doc.find_shape_object(mock_doc)
        self.assertEqual(found, body_obj)

    def test_doc_find_shape_object_part_fallback(self):
        """Test find_shape_object falls back to Part::Feature."""
        part_obj = MagicMock(Name="Part_Mock")
        part_obj.isDerivedFrom = lambda typeName: typeName == "Part::Feature"
        other_obj = MagicMock(Name="Other_Mock")
        other_obj.isDerivedFrom = lambda typeName: False
        mock_doc.Objects = [other_obj, part_obj]
        found = doc.find_shape_object(mock_doc)
        self.assertEqual(found, part_obj)

    def test_doc_find_shape_object_first_obj_fallback(self):
        """Test find_shape_object falls back to the first object."""
        other_obj1 = MagicMock(Name="Other1_Mock")
        other_obj1.isDerivedFrom = lambda typeName: False
        other_obj2 = MagicMock(Name="Other2_Mock")
        other_obj2.isDerivedFrom = lambda typeName: False
        mock_doc.Objects = [other_obj1, other_obj2]
        found = doc.find_shape_object(mock_doc)
        self.assertEqual(found, other_obj1)

    def test_doc_find_shape_object_no_objects(self):
        """Test find_shape_object returns None if document has no objects."""
        mock_doc.Objects = []
        found = doc.find_shape_object(mock_doc)
        self.assertIsNone(found)

    def test_doc_get_object_properties_found(self):
        """Test get_object_properties extracts existing properties."""
        setattr(mock_obj, "Diameter", "10 mm")
        setattr(mock_obj, "Length", "50 mm")
        params = doc.get_object_properties(mock_obj, ["Diameter", "Length"])
        # Expecting just the values, not tuples
        self.assertEqual(
            params,
            {
                "Diameter": ("10 mm", "App::PropertyString"),
                "Length": ("50 mm", "App::PropertyString"),
            },
        )
        mock_freecad.Console.PrintWarning.assert_not_called()

    def test_doc_get_object_properties_missing(self):
        """Test get_object_properties handles missing properties with warning."""
        # Re-import doc within the patch context to use the mocked FreeCAD
        import Path.Tool.shape.doc as doc_patched

        setattr(mock_obj, "Diameter", "10 mm")
        # Explicitly delete Height to ensure hasattr returns False for MagicMock
        if hasattr(mock_obj, "Height"):
            delattr(mock_obj, "Height")
        params = doc_patched.get_object_properties(mock_obj, ["Diameter", "Height"])
        # Expecting just the values, not tuples
        self.assertEqual(
            params,
            {
                "Diameter": ("10 mm", "App::PropertyString"),
                "Height": (None, "App::PropertyString"),
            },
        )  # Height is missing

    @patch("FreeCAD.openDocument")
    @patch("FreeCAD.getDocument")
    @patch("FreeCAD.closeDocument")
    def test_ShapeDocFromBytes(self, mock_close_doc, mock_get_doc, mock_open_doc):
        """Test ShapeDocFromBytes loads doc from a byte string."""
        content = b"fake_content"
        mock_opened_doc = MagicMock(Name="OpenedDoc_Mock")
        mock_get_doc.return_value = mock_opened_doc

        temp_file_path = None
        try:
            with doc.ShapeDocFromBytes(content=content) as temp_doc:
                # Verify temp file creation and content
                mock_open_doc.assert_called_once()
                temp_file_path = mock_open_doc.call_args[0][0]
                self.assertTrue(os.path.exists(temp_file_path))
                with open(temp_file_path, "rb") as f:
                    self.assertEqual(f.read(), content)

                self.assertEqual(temp_doc, mock_open_doc.return_value)

            # Verify cleanup after exiting the context
            mock_close_doc.assert_called_once_with(mock_open_doc.return_value.Name)
            self.assertFalse(os.path.exists(temp_file_path))

        finally:
            # Ensure cleanup even if test fails before assertion
            if temp_file_path and os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    @patch("FreeCAD.openDocument")
    @patch("FreeCAD.getDocument")
    @patch("FreeCAD.closeDocument")
    def test_ShapeDocFromBytes_open_exception(self, mock_close_doc, mock_get_doc, mock_open_doc):
        """Test ShapeDocFromBytes propagates exceptions and cleans up."""
        content = b"fake_content_exception"
        load_error = Exception("Fake load error")
        mock_open_doc.side_effect = load_error

        temp_file_path = None
        try:
            with self.assertRaises(Exception) as cm:
                with doc.ShapeDocFromBytes(content=content):
                    pass

            self.assertEqual(cm.exception, load_error)

            # Verify temp file was created before the exception
            mock_open_doc.assert_called_once()
            temp_file_path = mock_open_doc.call_args[0][0]
            self.assertTrue(os.path.exists(temp_file_path))
            with open(temp_file_path, "rb") as f:
                self.assertEqual(f.read(), content)

            mock_get_doc.assert_not_called()
            # closeDocument is called in __exit__ only if _doc is not None,
            # which it will be if openDocument failed.
            mock_close_doc.assert_not_called()

        finally:
            # Verify cleanup after exiting the context (even with exception)
            if temp_file_path and os.path.exists(temp_file_path):
                os.remove(temp_file_path)
            # Assert removal only if temp_file_path was set
            if temp_file_path:
                self.assertFalse(os.path.exists(temp_file_path))

    @patch("FreeCAD.openDocument")
    @patch("FreeCAD.getDocument")
    @patch("FreeCAD.closeDocument")
    def test_ShapeDocFromBytes_exit_cleans_up(self, mock_close_doc, mock_get_doc, mock_open_doc):
        """Test ShapeDocFromBytes __exit__ cleans up temp file."""
        content = b"fake_content_cleanup"
        mock_opened_doc = MagicMock(Name="OpenedDoc_Cleanup_Mock")
        mock_get_doc.return_value = mock_opened_doc

        temp_file_path = None
        try:
            with doc.ShapeDocFromBytes(content=content):
                mock_open_doc.assert_called_once()
                temp_file_path = mock_open_doc.call_args[0][0]
                self.assertTrue(os.path.exists(temp_file_path))
                with open(temp_file_path, "rb") as f:
                    self.assertEqual(f.read(), content)
                # No assertions on the returned doc here, focus is on cleanup
                pass  # Exit the context

            # Verify cleanup after exiting the context
            mock_close_doc.assert_called_once_with(mock_open_doc.return_value.Name)
            self.assertFalse(os.path.exists(temp_file_path))

        finally:
            # Ensure cleanup even if test fails before assertion
            if temp_file_path and os.path.exists(temp_file_path):
                os.remove(temp_file_path)


# Test execution
if __name__ == "__main__":
    unittest.main()



# ***************************************************************************
# *   Copyright (c) 2023 edi <edi271@a1.net>                                *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""
Provides the TechDraw PositionSectionView GuiCommand.
00.01 2021/03/17 C++ Basic version
00.02 2023/12/21 Option to select an edge and its corresponding vertex
"""

__title__ = "TechDrawTools.CommandPositionSectionView"
__author__ = "edi"
__url__ = "https://www.freecad.org"
__version__ = "00.02"
__date__ = "2023/12/21"

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui

import TechDrawTools.TDToolsUtil as Utils

class CommandPositionSectionView:
    """Orthogonally align a section view with its source view."""

    def __init__(self):
        """Initialize variables for the command that must exist at all times."""
        pass

    def GetResources(self):
        """Return a dictionary with data that will be used by the button or menu item."""
        return {'Pixmap': 'TechDraw_ExtensionPositionSectionView.svg',
                'Accel': "",
                'MenuText': QT_TRANSLATE_NOOP("TechDraw_PositionSectionView", "Position Section View"),
                'ToolTip': QT_TRANSLATE_NOOP("TechDraw_PositionSectionView",
                  "Aligns the selected section view with its source view orthogonally or the selected edge in the section view to the selected vertex in the base view")}

    def Activated(self):
        """Run the following code when the command is activated (button pressed)."""
        selection = Gui.Selection.getSelectionEx()
        if not selection:
            return
        if len(selection) == 1:
            if Utils.getSelView():
                sectionView = Utils.getSelView()
                if sectionView.TypeId == 'TechDraw::DrawViewSection':
                    baseView = sectionView.BaseView
                    if baseView.TypeId == "TechDraw::DrawProjGroupItem":
                        baseView = baseView.InList[0]
                    basePoint = App.Vector(baseView.X,baseView.Y,0.0)
                    sectionPoint = App.Vector(sectionView.X,sectionView.Y,0.0)
                    moveVector = sectionPoint.sub(basePoint)
                    if abs(moveVector.x) > abs(moveVector.y):
                        moveVector.x = 0.0
                    else:
                        moveVector.y = 0.0
                else:
                    return

        elif len(selection) == 2:
            if Utils.getSelEdges() and Utils.getSelVertexes(1,1):
                sectionEdge = Utils.getSelEdges()
                sectionDir = sectionEdge[0].Curve.Direction
                sectionSel = sectionEdge[0].Vertexes
                baseSel = Utils.getSelVertexes(1,1)
                sectionView = Utils.getSelView(0)
                baseView = Utils.getSelView(1)
                if baseView.TypeId == "TechDraw::DrawProjGroupItem":
                    baseView = baseView.InList[0]

                basePoint = baseSel[0].Point
                sectionPoint = sectionSel[0].Point
                Scale = baseView.Scale
                centerBase = App.Vector(baseView.X,baseView.Y,0)
                basePoint = centerBase+basePoint*Scale
                centerSection = App.Vector(sectionView.X,sectionView.Y,0)
                sectionPoint = centerSection+sectionPoint*Scale
                sectionPoint = self.getTrianglePoint(sectionPoint,sectionDir,basePoint)
                moveVector = sectionPoint.sub(basePoint)
            else:
                return

        sectionView.X = sectionView.X.Value-moveVector.x
        sectionView.Y = sectionView.Y.Value-moveVector.y

    def IsActive(self):
        """Return True when the command should be active or False when it should be disabled (greyed)."""
        if App.ActiveDocument:
            return Utils.havePage() and Utils.haveView()
        else:
            return False

    def getTrianglePoint(self,p1,dir,p2):
        '''
        Calculate the third vertex of a right triangle.

        Parameters:
        p1, p2 : vertices of the hypotenuse
        dir    : direction vector of one leg (kathete)

        Returns:
        p3 : the third vertex completing the right triangle
        '''
        a = -dir.y
        b = dir.x
        c1 = p1.x * a + p1.y * b
        c2 = -p2.x * b + p2.y * a
        ab = a * a + b * b
        x = (c1 * a - c2 * b) / ab
        y = (c2 * a + c1 * b) / ab
        return App.Vector(x,y,0.0)

# The command must be "registered" with a unique name by calling its class.
Gui.addCommand('TechDraw_ExtensionPositionSectionView', CommandPositionSectionView())




# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the ShapeString object."""
## @package shapestring
# \ingroup draftobjects
# \brief Provides the object code for the ShapeString object.

## \addtogroup draftobjects
# @{
import os
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import Part
from draftgeoutils import faces
from draftobjects.base import DraftObject
from draftutils import gui_utils
from draftutils.messages import _err, _log
from draftutils.translate import translate


class ShapeString(DraftObject):
    """The ShapeString object"""

    def __init__(self, obj):
        super().__init__(obj, "ShapeString")
        self.set_properties(obj)

    def set_properties(self, obj):
        """Add properties to the object and set them."""
        properties = obj.PropertiesList

        if "String" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Text string")
            obj.addProperty("App::PropertyString", "String", "Draft", _tip, locked=True)

        if "FontFile" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Font file name")
            obj.addProperty("App::PropertyFile", "FontFile", "Draft", _tip, locked=True)

        if "Size" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Height of text")
            obj.addProperty("App::PropertyLength", "Size", "Draft", _tip, locked=True)

        if "Justification" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Horizontal and vertical alignment")
            obj.addProperty("App::PropertyEnumeration", "Justification", "Draft", _tip, locked=True)
            obj.Justification = [
                "Top-Left",
                "Top-Center",
                "Top-Right",
                "Middle-Left",
                "Middle-Center",
                "Middle-Right",
                "Bottom-Left",
                "Bottom-Center",
                "Bottom-Right",
            ]
            obj.Justification = "Bottom-Left"

        if "JustificationReference" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Height reference used for justification")
            obj.addProperty(
                "App::PropertyEnumeration", "JustificationReference", "Draft", _tip, locked=True
            )
            obj.JustificationReference = ["Cap Height", "Shape Height"]
            obj.JustificationReference = "Cap Height"

        if "KeepLeftMargin" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Keep left margin and leading white space when justification is left",
            )
            obj.addProperty(
                "App::PropertyBool", "KeepLeftMargin", "Draft", _tip, locked=True
            ).KeepLeftMargin = False

        if "ScaleToSize" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Scale to ensure cap height is equal to size")
            obj.addProperty(
                "App::PropertyBool", "ScaleToSize", "Draft", _tip, locked=True
            ).ScaleToSize = True

        if "Tracking" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Inter-character spacing")
            obj.addProperty("App::PropertyDistance", "Tracking", "Draft", _tip, locked=True)

        if "ObliqueAngle" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Oblique (slant) angle")
            obj.addProperty("App::PropertyAngle", "ObliqueAngle", "Draft", _tip, locked=True)

        if "MakeFace" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Fill letters with faces")
            obj.addProperty(
                "App::PropertyBool", "MakeFace", "Draft", _tip, locked=True
            ).MakeFace = True

        if "Fuse" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Fuse faces if faces overlap, usually not required (can be very slow)",
            )
            obj.addProperty("App::PropertyBool", "Fuse", "Draft", _tip, locked=True).Fuse = False

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(
            obj, vp_module="view_shapestring", vp_class="ViewProviderShapeString"
        )
        if not hasattr(obj, "ObliqueAngle"):  # several more properties were added
            self.update_properties_1v0(obj)

    def update_properties_1v0(self, obj):
        """Update view properties."""
        old_tracking = obj.Tracking  # no need for obj.getTypeIdOfProperty("Tracking")
        obj.removeProperty("Tracking")
        self.set_properties(obj)
        obj.KeepLeftMargin = True
        obj.ScaleToSize = False
        obj.Tracking = old_tracking
        _log(
            "v1.0, "
            + obj.Name
            + ", "
            + "added 'Fuse', 'Justification', 'JustificationReference', 'KeepLeftMargin', "
            + "'ObliqueAngle' and 'ScaleToSize' properties"
        )
        _log("v1.0, " + obj.Name + ", changed 'Tracking' property type")

    def execute(self, obj):
        if self.props_changed_placement_only() or not obj.String or not obj.FontFile:
            obj.positionBySupport()
            self.props_changed_clear()
            return

        if obj.FontFile[0] == ".":
            # FontFile path relative to the FreeCAD file directory.
            font_file = os.path.join(os.path.dirname(obj.Document.FileName), obj.FontFile)
            # We need the absolute path to do some file checks.
            font_file = os.path.abspath(font_file)
        else:
            font_file = obj.FontFile

        # File checks:
        if not os.path.exists(font_file):
            _err(obj.Label + ": " + translate("draft", "Font file not found"))
            self.props_changed_clear()
            return
        if not os.path.isfile(font_file):
            _err(obj.Label + ": " + translate("draft", "Specified font file is not a file"))
            self.props_changed_clear()
            return
        if not os.path.splitext(font_file)[1].lower() in (".ttc", ".ttf", ".otf", ".pfb"):
            _err(obj.Label + ": " + translate("draft", "Specified font type is not supported"))
            self.props_changed_clear()
            return

        plm = obj.Placement
        fill = obj.MakeFace
        if fill is True:
            # Test a simple letter to know if we have a sticky font or not.
            # If the font is sticky change fill to `False`.
            # The 0.03 total area minimum is based on tests with:
            # 1CamBam_Stick_0.ttf and 1CamBam_Stick_0C.ttf.
            # See the make_faces function for more information.
            char = Part.makeWireString("L", font_file, 1, 0)[0]
            shapes = self.make_faces(char)  # char is list of wires
            if not shapes:
                fill = False
            else:
                # Depending on the font the size of char can be very small.
                # For the area check to make sense we need to use a scale factor.
                # https://github.com/FreeCAD/FreeCAD/issues/21501
                char_comp = Part.Compound(char)
                factor = 1 / char_comp.BoundBox.YLength
                fill = sum([shape.Area for shape in shapes]) > (0.03 / factor**2) and math.isclose(
                    char_comp.BoundBox.DiagonalLength,
                    Part.Compound(shapes).BoundBox.DiagonalLength,
                    rel_tol=1e-7,
                )

        chars = Part.makeWireString(obj.String, font_file, obj.Size, obj.Tracking)
        shapes = []

        for char in chars:
            if fill is False:
                shapes.extend(char)
            elif char:
                shapes.extend(self.make_faces(char))
        if shapes:
            if fill and obj.Fuse:
                ss_shape = shapes[0].fuse(shapes[1:])
                ss_shape = faces.concatenate(ss_shape)
                # Concatenate returns a Face or a Compound. We always
                # need a Compound as we use ss_shape.SubShapes later.
                if ss_shape.ShapeType == "Face":
                    ss_shape = Part.Compound([ss_shape])
            else:
                ss_shape = Part.Compound(shapes)
            cap_char = Part.makeWireString("M", font_file, obj.Size, obj.Tracking)[0]
            cap_height = Part.Compound(cap_char).BoundBox.YMax
            if obj.ScaleToSize:
                ss_shape.scale(obj.Size / cap_height)
                cap_height = obj.Size
            if obj.ObliqueAngle:
                if -80 <= obj.ObliqueAngle <= 80:
                    mtx = App.Matrix()
                    mtx.A12 = math.tan(math.radians(obj.ObliqueAngle))
                    ss_shape = ss_shape.transformGeometry(mtx)
                else:
                    wrn = (
                        translate(
                            "draft",
                            "ShapeString: oblique angle must be in the -80 to +80 degree range",
                        )
                        + "\n"
                    )
                    App.Console.PrintWarning(wrn)
            just_vec = self.justification_vector(
                ss_shape,
                cap_height,
                obj.Justification,
                obj.JustificationReference,
                obj.KeepLeftMargin,
            )
            shapes = ss_shape.SubShapes
            for shape in shapes:
                shape.translate(just_vec)
            obj.Shape = Part.Compound(shapes)
        else:
            App.Console.PrintWarning(translate("draft", "ShapeString: string has no wires") + "\n")

        obj.Placement = plm
        obj.positionBySupport()
        self.props_changed_clear()

    def onChanged(self, obj, prop):
        self.props_changed_store(prop)

    def justification_vector(
        self, ss_shape, cap_height, just, just_ref, keep_left_margin
    ):  # ss_shape is a compound
        box = ss_shape.optimalBoundingBox()
        if keep_left_margin is True and "Left" in just:
            vec = App.Vector(0, 0, 0)
        else:
            vec = App.Vector(
                -box.XMin, 0, 0
            )  # remove left margin caused by kerning and white space characters
        width = box.XLength
        if "Shape" in just_ref:
            vec = vec + App.Vector(0, -box.YMin, 0)
            height = box.YLength
        else:
            height = cap_height
        if "Top" in just:
            vec = vec + App.Vector(0, -height, 0)
        elif "Middle" in just:
            vec = vec + App.Vector(0, -height / 2, 0)
        if "Right" in just:
            vec = vec + App.Vector(-width, 0, 0)
        elif "Center" in just:
            vec = vec + App.Vector(-width / 2, 0, 0)
        return vec

    def make_faces(self, wireChar):
        wrn = translate("draft", "ShapeString: face creation failed for one character") + "\n"

        wirelist = []
        for w in wireChar:
            compEdges = Part.Compound(w.Edges)
            compEdges = compEdges.connectEdgesToWires()
            if compEdges.Wires[0].isClosed():
                wirelist.append(compEdges.Wires[0])

        if not wirelist:
            App.Console.PrintWarning(wrn)
            return []

        # Some test fonts:
        # https://raw.githubusercontent.com/FreeCAD/FPA/main/images/freecad_logo_official.svg
        #     https://evolventa.github.io/
        #     not a problem font, but it is used by FreeCAD
        # https://forum.freecad.org/viewtopic.php?t=57774
        #     https://www.dafont.com/mutlu-ornamental.font
        # https://forum.freecad.org/viewtopic.php?t=65110&p=559810#p559886
        #     http://www.atelier-des-fougeres.fr/Cambam/Aide/Plugins/stickfonts.html

        # 1CamBam_Stick_0.ttf is actually not a stick font.

        # FaceMakerBullseye:
        #     1CamBam_Stick_0.ttf face validation problem with A, E, F, H, K, R, Y and y.
        # FaceMakerCheese:
        #     1CamBam_Stick_0.ttf face creation problem with: A, E, F, H, Q, R, e and y.
        #     All fonts: face creation problem in case of double-nested wires f.e. with: ©.
        # FaceMakerSimple:
        #     All fonts: overlapping faces in case of nested wires f.e. with: O.
        try:
            # print("try Bullseye")
            faces = Part.makeFace(wirelist, "Part::FaceMakerBullseye").Faces
            for face in faces:
                face.validate()
        except Part.OCCError:
            try:
                # print("try Cheese")
                faces = Part.makeFace(wirelist, "Part::FaceMakerCheese").Faces
                for face in faces:
                    face.validate()
            except Part.OCCError:
                try:
                    # print("try Simple")
                    faces = Part.makeFace(wirelist, "Part::FaceMakerSimple").Faces
                    for face in faces:
                        face.validate()
                except Part.OCCError:
                    App.Console.PrintWarning(wrn)
                    return []

        for face in faces:
            try:
                # some fonts fail here
                if face.normalAt(0, 0).z < 0:  # Does not seem to occur for FaceMakerBullseye.
                    face.reverse()
            except Exception:
                pass

        return faces


# Alias for compatibility with v0.18 and earlier
_ShapeString = ShapeString

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2021 FreeCAD Developers                                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Run this file to create a standard test document for Draft objects.

Use it as input for the program executable.

::

    freecad draft_test_objects.py

Or load it as a module and use the defined function.

>>> import drafttests.draft_test_objects as dt
>>> dt.create_test_file()
"""

## @package draft_test_objects
# \ingroup drafttests
# \brief Run this file to create a standard test document for Draft objects.

## \addtogroup drafttests
# @{

import datetime

import FreeCAD as App
import Part
import Draft
from FreeCAD import Vector
from draftutils.messages import _msg, _wrn


if App.GuiUp:
    import FreeCADGui as Gui


def _set_text(text_list, position):
    """Set a text annotation with properties."""
    obj = App.ActiveDocument.addObject("App::Annotation", "Annotation")
    obj.LabelText = text_list
    obj.Position = position

    if App.GuiUp:
        obj.ViewObject.DisplayMode = "World"
        obj.ViewObject.FontSize = 75
        obj.ViewObject.TextColor = (0.0, 0.0, 0.0)


def _create_frame(doc=None):
    """Draw a frame with information on the version of the software."""
    if not doc:
        doc = App.activeDocument()
    if not doc:
        doc = App.newDocument()

    version = App.Version()
    now = datetime.datetime.now().strftime("%Y/%m/%dT%H:%M:%S")

    _text = [
        "Draft test file",
        "Created: {}".format(now),
        "\n",
        "Version: " + ".".join(version[0:3]),
        "Release: " + " ".join(version[3:5]),
        "Branch: " + " ".join(version[5:]),
    ]

    record = doc.addObject("App::Annotation", "Description")
    record.LabelText = _text
    record.Position = Vector(0, -1000, 0)

    if App.GuiUp:
        record.ViewObject.DisplayMode = "World"
        record.ViewObject.FontSize = 400
        record.ViewObject.TextColor = (0.0, 0.0, 0.0)

    p1 = Vector(-1000, -4000, 0)
    p2 = Vector(17000, -4000, 0)
    p3 = Vector(17000, 8000, 0)
    p4 = Vector(-1000, 8000, 0)

    poly = Part.makePolygon([p1, p2, p3, p4, p1])
    frame = doc.addObject("Part::Feature", "Frame")
    frame.Shape = poly


def _create_objects(doc=None, font_file=None, hatch_file=None, hatch_name=None):
    """Create the objects of the test file."""
    if not doc:
        doc = App.activeDocument()
    if not doc:
        doc = App.newDocument()

    # Drafting ##############################################################

    # Line
    _msg(16 * "-")
    _msg("Line")
    Draft.make_line(Vector(0, 0, 0), Vector(500, 500, 0))
    _set_text(["Line"], Vector(0, -200, 0))

    # Wire
    _msg(16 * "-")
    _msg("Wire")
    Draft.make_wire([Vector(1000, 0, 0), Vector(1500, 250, 0), Vector(1500, 500, 0)])
    _set_text(["Wire"], Vector(1000, -200, 0))

    # Fillet
    _msg(16 * "-")
    _msg("Fillet")
    line_1 = Draft.make_line(Vector(2000, 0, 0), Vector(2000, 500, 0))
    line_2 = Draft.make_line(Vector(2000, 500, 0), Vector(2500, 500, 0))
    if App.GuiUp:
        line_1.ViewObject.DrawStyle = "Dotted"
        line_2.ViewObject.DrawStyle = "Dotted"
    doc.recompute()
    Draft.make_fillet([line_1, line_2], 400)
    _set_text(["Fillet"], Vector(2000, -200, 0))

    # Circular arc
    _msg(16 * "-")
    _msg("Circular arc")
    arc = Draft.make_circle(250, startangle=90, endangle=270)
    arc.Placement.Base = Vector(3250, 250, 0)
    _set_text(["Circular arc"], Vector(3000, -200, 0))

    # Circular arc 3 points
    _msg(16 * "-")
    _msg("Circular arc 3 points")
    Draft.make_arc_3points([Vector(4250, 0, 0), Vector(4000, 250, 0), Vector(4250, 500, 0)])
    _set_text(["Circular arc 3 points"], Vector(4000, -200, 0))

    # Circle
    _msg(16 * "-")
    _msg("Circle")
    circle = Draft.make_circle(250)
    circle.Placement.Base = Vector(5250, 250, 0)
    _set_text(["Circle"], Vector(5000, -200, 0))

    # Ellipse
    _msg(16 * "-")
    _msg("Ellipse")
    ellipse = Draft.make_ellipse(250, 150)
    ellipse.Placement.Base = Vector(6250, 150, 0)
    _set_text(["Ellipse"], Vector(6000, -200, 0))

    # Rectangle
    _msg(16 * "-")
    _msg("Rectangle")
    rectangle = Draft.make_rectangle(500, 300, 0)
    rectangle.Placement.Base = Vector(7000, 0, 0)
    _set_text(["Rectangle"], Vector(7000, -200, 0))

    # Polygon
    _msg(16 * "-")
    _msg("Polygon")
    polygon = Draft.make_polygon(5, 250)
    polygon.Placement.Base = Vector(8250, 250, 0)
    _set_text(["Polygon"], Vector(8000, -200, 0))

    # BSpline
    _msg(16 * "-")
    _msg("BSpline")
    Draft.make_bspline(
        [Vector(9000, 0, 0), Vector(9100, 200, 0), Vector(9400, 300, 0), Vector(9500, 500, 0)]
    )
    _set_text(["BSpline"], Vector(9000, -200, 0))

    # Cubic bezier
    _msg(16 * "-")
    _msg("Cubic bezier")
    Draft.make_bezcurve(
        [Vector(10000, 0, 0), Vector(10000, 500, 0), Vector(10500, 0, 0), Vector(10500, 500, 0)],
        degree=3,
    )
    _set_text(["Cubic bezier"], Vector(10000, -200, 0))

    # N-degree bezier
    _msg(16 * "-")
    _msg("N-degree bezier")
    Draft.make_bezcurve(
        [
            Vector(11000, 0, 0),
            Vector(11100, 400, 0),
            Vector(11250, 250, 0),
            Vector(11400, 100, 0),
            Vector(11500, 500, 0),
        ]
    )
    _set_text(["N-degree bezier"], Vector(11000, -200, 0))

    # Point
    _msg(16 * "-")
    _msg("Point")
    point = Draft.make_point(12000, 0, 0)
    if App.GuiUp:
        point.ViewObject.PointSize = 10
    _set_text(["Point"], Vector(12000, -200, 0))

    # Facebinder
    _msg(16 * "-")
    _msg("Facebinder")
    box = doc.addObject("Part::Box", "Box")
    box.Length = 200
    box.Width = 500
    box.Height = 100
    box.Placement.Base = Vector(13000, 0, 0)
    if App.GuiUp:
        box.ViewObject.Visibility = False
    facebinder = Draft.make_facebinder([(box, ("Face1", "Face3", "Face6"))])
    facebinder.Extrusion = 10
    _set_text(["Facebinder"], Vector(13000, -200, 0))

    # Shapestring
    _msg(16 * "-")
    _msg("Shapestring")
    try:
        shape_string = Draft.make_shapestring("Testing", font_file, 100)
        shape_string.Placement.Base = Vector(14000, 0)
    except Exception:
        _wrn("Shapestring could not be created")
        _wrn("Possible cause: the font file may not exist")
        _wrn(font_file)
    _set_text(["Shapestring"], Vector(14000, -200, 0))

    # Hatch
    _msg(16 * "-")
    _msg("Hatch")
    rectangle = Draft.make_rectangle(500, 300, 0)
    rectangle.Placement.Base = Vector(15000, 0, 0)
    rectangle.MakeFace = True
    if App.GuiUp:
        rectangle.ViewObject.Visibility = False
    doc.recompute()
    try:
        Draft.make_hatch(rectangle, hatch_file, hatch_name, scale=10, rotation=45)
    except Exception:
        _wrn("Hatch could not be created")
        _wrn("Possible cause: the hatch file may not exist")
        _wrn(hatch_file)
    _set_text(["Hatch"], Vector(15000, -200, 0))

    # Annotation ############################################################

    # Text
    _msg(16 * "-")
    _msg("Text")
    text = Draft.make_text(["Testing", "text"], Vector(0, 2100, 0))
    if App.GuiUp:
        text.ViewObject.FontSize = 100
    _set_text(["Text"], Vector(0, 1800, 0))

    # Linear dimension
    _msg(16 * "-")
    _msg("Linear dimension")
    dimension = Draft.make_linear_dimension(
        Vector(1500, 2000, 0), Vector(1500, 2400, 0), Vector(1000, 2200, 0)
    )
    if App.GuiUp:
        dimension.ViewObject.ArrowSizeStart = 15
        dimension.ViewObject.ArrowSizeEnd = 15
        dimension.ViewObject.ExtLines = 0
        dimension.ViewObject.ExtOvershoot = 50
        dimension.ViewObject.DimOvershoot = 25
        dimension.ViewObject.FontSize = 50
        dimension.ViewObject.Decimals = 1
        dimension.ViewObject.ShowUnit = False

    line = Draft.make_wire([Vector(1500, 2600, 0), Vector(1500, 3000, 0)])
    doc.recompute()
    dimension = Draft.make_linear_dimension_obj(line, 1, 2, Vector(1000, 2800, 0))
    if App.GuiUp:
        dimension.ViewObject.ArrowSizeStart = 15
        dimension.ViewObject.ArrowSizeEnd = 15
        dimension.ViewObject.ArrowTypeStart = "Arrow"
        dimension.ViewObject.ArrowTypeEnd = "Arrow"
        dimension.ViewObject.ExtLines = -50
        dimension.ViewObject.ExtOvershoot = 50
        dimension.ViewObject.DimOvershoot = 25
        dimension.ViewObject.FontSize = 50
        dimension.ViewObject.Decimals = 1
        dimension.ViewObject.ShowUnit = False

    _set_text(["Dimension"], Vector(1000, 1800, 0))

    # Radius and diameter dimension
    _msg(16 * "-")
    _msg("Radius and diameter dimension")
    circle = Draft.make_circle(200)
    circle.Placement.Base = Vector(2200, 2200, 0)
    circle.MakeFace = False
    doc.recompute()
    dimension = Draft.make_radial_dimension_obj(circle, 1, "radius", Vector(2300, 2300, 0))
    if App.GuiUp:
        dimension.ViewObject.ArrowSizeStart = 15
        dimension.ViewObject.ArrowSizeEnd = 15
        dimension.ViewObject.FontSize = 50
        dimension.ViewObject.Decimals = 1
        dimension.ViewObject.ShowUnit = False

    circle = Draft.make_circle(200)
    circle.Placement.Base = Vector(2200, 2800, 0)
    circle.MakeFace = False
    doc.recompute()
    dimension = Draft.make_radial_dimension_obj(circle, 1, "diameter", Vector(2300, 2900, 0))
    if App.GuiUp:
        dimension.ViewObject.ArrowSizeStart = 15
        dimension.ViewObject.ArrowSizeEnd = 15
        dimension.ViewObject.FontSize = 50
        dimension.ViewObject.Decimals = 1
        dimension.ViewObject.ShowUnit = False
    _set_text(["Radius dimension", "Diameter dimension"], Vector(2000, 1800, 0))

    # Angular dimension
    _msg(16 * "-")
    _msg("Angular dimension")
    Draft.make_line(Vector(3000, 2000, 0), Vector(3500, 2000, 0))
    Draft.make_line(Vector(3000, 2000, 0), Vector(3500, 2500, 0))
    dimension = Draft.make_angular_dimension(Vector(3000, 2000, 0), [0, 45], Vector(3250, 2250, 0))
    if App.GuiUp:
        dimension.ViewObject.ArrowSizeStart = 15
        dimension.ViewObject.ArrowSizeEnd = 15
        dimension.ViewObject.FontSize = 50
        dimension.ViewObject.Decimals = 1
    _set_text(["Angle dimension"], Vector(3000, 1800, 0))

    # Label
    _msg(16 * "-")
    _msg("Label")
    place = App.Placement(Vector(4250, 2250, 0), App.Rotation())
    label = Draft.make_label(
        target_point=Vector(4000, 2000, 0),
        placement=place,
        custom_text="Example label",
        distance=-100,
    )
    label.Text = "Testing"
    if App.GuiUp:
        label.ViewObject.ArrowSizeStart = 15
        label.ViewObject.FontSize = 50
    _set_text(["Label"], Vector(4000, 1800, 0))

    # Array #################################################################

    # Orthogonal array
    _msg(16 * "-")
    _msg("Orthogonal array")
    rectangle = Draft.make_rectangle(100, 50)
    rectangle.Placement.Base = Vector(0, 4000, 0)
    if App.GuiUp:
        rectangle.ViewObject.Visibility = False
    doc.recompute()
    Draft.make_ortho_array(
        rectangle, Vector(200, 0, 0), Vector(0, 150, 0), Vector(0, 0, 0), 3, 2, 1, use_link=False
    )
    _set_text(["Orthogonal array"], Vector(0, 3800, 0))

    # Orthogonal link array
    _msg(16 * "-")
    _msg("Orthogonal link array")
    rectangle = Draft.make_rectangle(50, 50)
    rectangle.Placement.Base = Vector(1000, 4000, 0)
    if App.GuiUp:
        rectangle.ViewObject.Visibility = False
    doc.recompute()
    Draft.make_ortho_array(
        rectangle, Vector(200, 0, 0), Vector(0, 150, 0), Vector(0, 0, 0), 3, 2, 1, use_link=True
    )
    _set_text(["Orthogonal link array"], Vector(1000, 3800, 0))

    # Polar array
    _msg(16 * "-")
    _msg("Polar array")
    wire = Draft.make_wire([Vector(2000, 4050, 0), Vector(2000, 4000, 0), Vector(2100, 4000, 0)])
    if App.GuiUp:
        wire.ViewObject.Visibility = False
    doc.recompute()
    Draft.make_polar_array(wire, 4, 90, Vector(2000, 4250, 0), use_link=False)
    _set_text(["Polar array"], Vector(2000, 3800, 0))

    # Polar link array
    _msg(16 * "-")
    _msg("Polar link array")
    wire = Draft.make_wire([Vector(3000, 4050, 0), Vector(3000, 4000, 0), Vector(3050, 4000, 0)])
    if App.GuiUp:
        wire.ViewObject.Visibility = False
    doc.recompute()
    Draft.make_polar_array(wire, 4, 90, Vector(3000, 4250, 0), use_link=True)
    _set_text(["Polar link array"], Vector(3000, 3800, 0))

    # Circular array
    _msg(16 * "-")
    _msg("Circular array")
    polygon = Draft.make_polygon(3, 30)
    polygon.Placement.Base = Vector(4250, 4250, 0)
    if App.GuiUp:
        polygon.ViewObject.Visibility = False
    doc.recompute()
    Draft.make_circular_array(
        polygon, 110, 100, 3, 1, Vector(0, 0, 1), Vector(4250, 4250, 0), use_link=False
    )
    _set_text(["Circular array"], Vector(4000, 3800, 0))

    # Circular link array
    _msg(16 * "-")
    _msg("Circular link array")
    polygon = Draft.make_polygon(4, 30)
    polygon.Placement.Base = Vector(5250, 4250, 0)
    if App.GuiUp:
        polygon.ViewObject.Visibility = False
    doc.recompute()
    Draft.make_circular_array(
        polygon, 110, 100, 3, 1, Vector(0, 0, 1), Vector(5250, 4250, 0), use_link=True
    )
    _set_text(["Circular link array"], Vector(5000, 3800, 0))

    # Path array
    _msg(16 * "-")
    _msg("Path array")
    polygon = Draft.make_polygon(3, 30)
    polygon.Placement.Base = Vector(6000, 4000, 0)
    if App.GuiUp:
        polygon.ViewObject.Visibility = False
    spline = Draft.make_bspline(
        [Vector(6000, 4000, 0), Vector(6100, 4200, 0), Vector(6400, 4300, 0), Vector(6500, 4500, 0)]
    )
    doc.recompute()
    Draft.make_path_array(polygon, spline, 5, use_link=False)
    _set_text(["Path array"], Vector(6000, 3800, 0))

    # Path link array
    _msg(16 * "-")
    _msg("Path link array")
    polygon = Draft.make_polygon(4, 30)
    polygon.Placement.Base = Vector(7000, 4000, 0)
    if App.GuiUp:
        polygon.ViewObject.Visibility = False
    spline = Draft.make_bspline(
        [Vector(7000, 4000, 0), Vector(7100, 4200, 0), Vector(7400, 4300, 0), Vector(7500, 4500, 0)]
    )
    doc.recompute()
    Draft.make_path_array(polygon, spline, 5, use_link=True)
    _set_text(["Path link array"], Vector(7000, 3800, 0))

    # Point array
    _msg(16 * "-")
    _msg("Point array")
    polygon = Draft.make_polygon(3, 30)
    polygon.Placement.Base = Vector(8000, 4000, 0)
    point_1 = Draft.make_point(8030, 4030, 0)
    point_2 = Draft.make_point(8030, 4250, 0)
    point_3 = Draft.make_point(8470, 4250, 0)
    point_4 = Draft.make_point(8470, 4470, 0)
    add_list, delete_list = Draft.upgrade([point_1, point_2, point_3, point_4])
    compound = add_list[0]
    if App.GuiUp:
        compound.ViewObject.PointSize = 5
    doc.recompute()
    Draft.make_point_array(polygon, compound, use_link=False)
    _set_text(["Point array"], Vector(8000, 3800, 0))

    # Point link array
    _msg(16 * "-")
    _msg("Point link array")
    polygon = Draft.make_polygon(4, 30)
    polygon.Placement.Base = Vector(9000, 4000, 0)
    point_1 = Draft.make_point(9030, 4030, 0)
    point_2 = Draft.make_point(9030, 4250, 0)
    point_3 = Draft.make_point(9470, 4250, 0)
    point_4 = Draft.make_point(9470, 4470, 0)
    add_list, delete_list = Draft.upgrade([point_1, point_2, point_3, point_4])
    compound = add_list[0]
    if App.GuiUp:
        compound.ViewObject.PointSize = 5
    doc.recompute()
    Draft.make_point_array(polygon, compound, use_link=True)
    _set_text(["Point link array"], Vector(9000, 3800, 0))

    # Miscellaneous #########################################################

    # Mirror
    _msg(16 * "-")
    _msg("Mirror")
    wire = Draft.make_wire([Vector(0, 6000, 0), Vector(150, 6200, 0), Vector(500, 6000, 0)])
    Draft.mirror(wire, Vector(0, 6250, 0), Vector(500, 6250, 0))
    _set_text(["Mirror"], Vector(0, 5800, 0))

    # Clone
    _msg(16 * "-")
    _msg("Clone")
    wire = Draft.make_wire([Vector(1000, 6000, 0), Vector(1150, 6200, 0), Vector(1500, 6000, 0)])
    Draft.make_clone(wire, Vector(0, 300, 0))
    _set_text(["Clone"], Vector(1000, 5800, 0))

    # Shape2DView
    _msg(16 * "-")
    _msg("Shape2DView")
    place = App.Placement(Vector(2000, 6000, 0), App.Rotation(Vector(0, 0, 1), Vector(1, 2, 3)))
    box = doc.addObject("Part::Box", "Box")
    box.Length = 200
    box.Width = 500
    box.Height = 100
    box.Placement = place
    if App.GuiUp:
        box.ViewObject.Visibility = False
    Draft.make_shape2dview(box)
    _set_text(["Shape2DView"], Vector(2000, 5800, 0))

    # WorkingPlaneProxy
    _msg(16 * "-")
    _msg("WorkingPlaneProxy")
    place = App.Placement(Vector(3250, 6250, 0), App.Rotation())
    proxy = Draft.make_workingplaneproxy(place)
    if App.GuiUp:
        proxy.ViewObject.DisplaySize = 500
        proxy.ViewObject.ArrowSize = 50
    _set_text(["WorkingPlaneProxy"], Vector(3000, 5800, 0))

    # Layer
    _msg(16 * "-")
    _msg("Layer")
    layer = Draft.make_layer(
        "Custom layer",
        line_color=(0.33, 0.0, 0.49),
        shape_color=(0.56, 0.89, 0.56),
        line_width=4,
        draw_style="Solid",
        transparency=50,
    )
    box = doc.addObject("Part::Box", "Box")
    box.Length = 200
    box.Width = 500
    box.Height = 100
    box.Placement.Base = Vector(4000, 6000, 0)
    sphere = doc.addObject("Part::Sphere", "Sphere")
    sphere.Radius = 100
    sphere.Placement.Base = Vector(4400, 6250, 0)
    layer.Proxy.addObject(layer, box)
    layer.Proxy.addObject(layer, sphere)
    _set_text(["Layer"], Vector(4000, 5800, 0))

    doc.recompute()


def create_test_file(
    font_file=App.getResourceDir() + "Mod/TechDraw/Resources/fonts/osifont-lgpl3fe.ttf",
    hatch_file=App.getResourceDir() + "Mod/TechDraw/PAT/FCPAT.pat",
    hatch_name="Horizontal5",
):
    """Create a complete test file of Draft objects.

    It draws a frame with information on the software used to create
    the test document, and fills it with every object that can be created.

    Parameters
    ----------
    font_file: str, optional
        It defaults to `App.getHomePath()+"data/Mod/TechDraw/Resources/fonts/osifont-lgpl3fe.ttf"`
        It is the full path of a font file to be used to create a `Draft ShapeString`.
        If the file is not found, this object is not created.

    hatch_file: str, optional
        It defaults to `App.getHomePath()+"data/Mod/TechDraw/PAT/FCPAT.pat"`
        It is the full path of a PAT file to be used to create a `Draft Hatch`.
        If the file is not found, this object is not created.

    hatch_name: str, optional
        It defaults to `"Horizontal5"`
        It is the name of a hatch pattern in the hatch_file.

    Returns
    -------
    App::Document
        A reference to the test document that was created.
    """
    _msg(16 * "-")
    _msg("If the units tests fail, this script may fail as well")
    doc = App.newDocument()

    _create_frame(doc=doc)
    _create_objects(doc=doc, font_file=font_file, hatch_file=hatch_file, hatch_name=hatch_name)

    if App.GuiUp:
        Gui.runCommand("Std_ViewFitAll")
        Gui.Selection.clearSelection()

    return doc


## @}


if __name__ == "__main__":
    create_test_file()



# ***************************************************************************
# *   Copyright (c) 2018 Przemo Firszt <przemo@firszt.eu>                   *
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

# Unit test for the FEM module
# to get the right order import as is used
from femtest.app.test_femimport import TestFemImport as FemTest01
from femtest.app.test_common import TestFemCommon as FemTest02
from femtest.app.test_object import TestObjectCreate as FemTest03
from femtest.app.test_object import TestObjectType as FemTest04
from femtest.app.test_open import TestObjectOpen as FemTest05
from femtest.app.test_material import TestMaterialUnits as FemTest06
from femtest.app.test_mesh import TestMeshCommon as FemTest07
from femtest.app.test_mesh import TestMeshEleTetra10 as FemTest08
from femtest.app.test_mesh import TestMeshGroups as FemTest09
from femtest.app.test_result import TestResult as FemTest10
from femtest.app.test_ccxtools import TestCcxTools as FemTest11
from femtest.app.test_solver_elmer import TestSolverElmer as FemTest13
from femtest.app.test_solver_z88 import TestSolverZ88 as FemTest14

# dummy usage to get flake8 and lgtm quiet
False if FemTest01.__name__ else True
False if FemTest02.__name__ else True
False if FemTest03.__name__ else True
False if FemTest04.__name__ else True
False if FemTest05.__name__ else True
False if FemTest06.__name__ else True
False if FemTest07.__name__ else True
False if FemTest08.__name__ else True
False if FemTest09.__name__ else True
False if FemTest10.__name__ else True
False if FemTest11.__name__ else True
False if FemTest13.__name__ else True
False if FemTest14.__name__ else True



# /******************************************************************************
# *   Copyright (c) 2020 Werner Mayer <wmayer[at]users.sourceforge.net>        *
# *                                                                            *
# *   This file is part of the FreeCAD CAx development system.                 *
# *                                                                            *
# *   This library is free software; you can redistribute it and/or            *
# *   modify it under the terms of the GNU Library General Public              *
# *   License as published by the Free Software Foundation; either             *
# *   version 2 of the License, or (at your option) any later version.         *
# *                                                                            *
# *   This library  is distributed in the hope that it will be useful,         *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
# *   GNU Library General Public License for more details.                     *
# *                                                                            *
# *   You should have received a copy of the GNU Library General Public        *
# *   License along with this library; see the file COPYING.LIB. If not,       *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,            *
# *   Suite 330, Boston, MA  02111-1307, USA                                   *
# *                                                                            *
# ******************************************************************************/


import FreeCAD as App
import FreeCADGui as Gui
from PySide import QtGui
from freecad.utils import get_python_exe


class RemoteDebugger:
    def __init__(self, parent=None):
        ui = App.getHomePath() + "Ext/freecad/gui/RemoteDebugger.ui"
        self.dialog = Gui.PySideUic.loadUi(ui)
        self.dialog.buttonBox.accepted.connect(self.accept)
        self.dialog.buttonBox.rejected.connect(self.reject)

        self.prefs = App.ParamGet("User parameter:BaseApp/Macro/Debugger")
        index = self.prefs.GetInt("TabIndex", 0)
        self.dialog.tabWidget.setCurrentIndex(index)
        address = self.prefs.GetString("VSCodeAddress", "localhost")
        port = self.prefs.GetInt("VSCodePort", 5678)
        self.dialog.lineEditAddress.setText(address)
        self.dialog.spinBoxPort.setValue(port)

    def accept(self):
        try:
            index = self.dialog.tabWidget.currentIndex()
            self.prefs.SetInt("TabIndex", index)

            if index == 0:  # winpdb
                passwd = self.dialog.lineEditPassword.text()

                import rpdb2

                rpdb2.start_embedded_debugger(passwd, timeout=30)

            elif index == 1:  # VS code
                address = self.dialog.lineEditAddress.text()
                port = self.dialog.spinBoxPort.value()
                self.prefs.SetString("VSCodeAddress", address)
                self.prefs.SetInt("VSCodePort", port)

                import debugpy

                # get_python_exe is needed because debugpy needs a python interpreter to work.
                # It does not have to be FC embedded interpreter.
                # By default it attempts to use Freecad's PID mistaking it for python.
                # https://github.com/microsoft/debugpy/issues/262
                debugpy.configure(python=get_python_exe())
                debugpy.listen((address, port))
                debugpy.wait_for_client()

        except Exception as e:
            QtGui.QMessageBox.warning(self.dialog, "Failed to attach", str(e))

        self.dialog.accept()

    def reject(self):
        self.dialog.reject()

    def exec_(self):
        self.dialog.exec_()


def attachToRemoteDebugger():
    dlg = RemoteDebugger(Gui.getMainWindow())
    dlg.exec_()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the Draft Workbench, DXF import and export tests."""

## @package test_dxf
# \ingroup drafttests
# \brief Unit tests for the Draft Workbench, DXF import and export tests.

## \addtogroup drafttests
# @{

import os

import FreeCAD as App
import Draft
from drafttests import auxiliary as aux
from drafttests import test_base
from draftutils.messages import _msg
import importDXF


class DraftDXF(test_base.DraftTestCaseDoc):
    """Test reading and writing of DXF files with Draft."""

    def test_read_dxf_Issue24314(self):
        """Verify that reading a DXF file does not leave pending Python error states"""

        file = "Mod/Draft/drafttests/Issue24314.dxf"
        in_file = os.path.join(App.getHomePath(), file)
        _msg("  file={}".format(in_file))
        _msg("  exists={}".format(os.path.exists(in_file)))

        hGrp = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")

        # Set options, doing our best to restore them:
        wasShowDialog = hGrp.GetBool("dxfShowDialog", True)
        wasUseLegacyImporter = hGrp.GetBool("dxfUseLegacyImporter", False)
        wasUseLayers = hGrp.GetBool("dxfUseDraftVisGroups", True)
        wasImportMode = hGrp.GetInt("DxfImportMode", 2)
        wasCreateSketch = hGrp.GetBool("dxfCreateSketch", False)
        wasImportAnonymousBlocks = hGrp.GetBool("dxfstarblocks", False)

        doc = None
        try:
            # disable Preferences dialog in gui mode (avoids popup prompt to user)
            hGrp.SetBool("dxfShowDialog", False)
            # Use the new C++ importer -- that's where the bug was
            hGrp.SetBool("dxfUseLegacyImporter", False)
            # Preserve the DXF layers (makes the checking of document contents easier)
            hGrp.SetBool("dxfUseDraftVisGroups", True)
            # create simple part shapes (2 params)
            # This is required to display the bug because creation of Draft objects clears out the
            # pending exception this test is looking for, whereas creation of the simple shape object
            # actually throws on the pending exception so the entity is absent from the document.
            hGrp.SetInt("DxfImportMode", 2)
            hGrp.SetBool("dxfCreateSketch", False)
            hGrp.SetBool("dxfstarblocks", False)
            doc = importDXF.open(in_file)
            # This doc should have 3 objects: The Layers container, the DXF layer called 0, and one Line
            self.assertEqual(len(doc.Objects), 3)
        finally:
            hGrp.SetBool("dxfShowDialog", wasShowDialog)
            hGrp.SetBool("dxfUseLegacyImporter", wasUseLegacyImporter)
            hGrp.SetBool("dxfUseDraftVisGroups", wasUseLayers)
            hGrp.SetInt("DxfImportMode", wasImportMode)
            hGrp.SetBool("dxfCreateSketch", wasCreateSketch)
            hGrp.SetBool("dxfstarblocks", wasImportAnonymousBlocks)
            if doc:
                App.closeDocument(doc.Name)

    def test_export_dxf(self):
        """Create some figures and export them to a DXF file."""
        operation = "importDXF.export"
        _msg("  Test '{}'".format(operation))

        file = "Mod/Draft/drafttest/out_test.dxf"
        out_file = os.path.join(App.getResourceDir(), file)
        _msg("  file={}".format(out_file))
        _msg("  exists={}".format(os.path.exists(out_file)))

        obj = aux.fake_function(out_file)
        self.assertTrue(obj, "'{}' failed".format(operation))


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The Bim Sketch command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_Sketch:

    def GetResources(self):
        return {
            "Pixmap": "Sketch",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Sketch", "Sketch"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Sketch", "Creates a new sketch in the current working plane"
            ),
            "Accel": "S,K",
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        import WorkingPlane
        from draftutils import params
        from draftutils import utils
        from FreeCAD import Units

        wp = WorkingPlane.get_working_plane()  # also updates the grid
        sk = FreeCAD.ActiveDocument.addObject("Sketcher::SketchObject", "Sketch")
        sk.Placement = wp.get_placement()
        sk.MapMode = "Deactivated"

        if not params.get_param("BIMSketchPlacementOnly", path="Mod/BIM"):
            sk.ViewObject.LineWidth = params.get_param_view("DefaultShapeLineWidth")
            sk.ViewObject.PointSize = params.get_param_view("DefaultShapePointSize")
            sk.ViewObject.AutoColor = False
            sk.ViewObject.LineColor = params.get_param_view("DefaultShapeLineColor")
            sk.ViewObject.PointColor = params.get_param_view("DefaultShapeVertexColor")
            sk.ViewObject.ShapeAppearance = [utils.get_view_material()]
            if getattr(FreeCADGui, "Snapper", None) and FreeCADGui.Snapper.grid.Visible:
                sk.ViewObject.GridSize = Units.Quantity(params.get_param("gridSpacing"))
                sk.ViewObject.ShowGrid = True

        FreeCADGui.ActiveDocument.setEdit(sk.Name)
        FreeCADGui.activateWorkbench("SketcherWorkbench")


FreeCADGui.addCommand("BIM_Sketch", BIM_Sketch())



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM mesh gmsh document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package mesh_gmsh
#  \ingroup FEM
#  \brief mesh gmsh object

from FreeCAD import Base
from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class MeshGmsh(base_fempythonobject.BaseFemPythonObject):
    """
    A Fem::FemMeshObject python type, add Gmsh specific properties
    """

    Type = "Fem::FemMeshGmsh"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyLinkList",
                name="MeshBoundaryLayerList",
                group="Base",
                doc="Mesh boundaries need inflation layers",
                value=[],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLinkList",
                name="MeshRegionList",
                group="Base",
                doc="Mesh refinments of the mesh",
                value=[],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLinkList",
                name="MeshGroupList",
                group="Base",
                doc="Mesh groups of the mesh",
                value=[],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="CharacteristicLengthMax",
                group="Mesh Parameters",
                doc="Max mesh element size (0.0 means infinity)",
                value=0.0,  # will be 1e+22
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="CharacteristicLengthMin",
                group="Mesh Parameters",
                doc="Min mesh element size",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="ElementDimension",
                group="Mesh Parameters",
                doc="Dimension of mesh elements ('From Shape': according ShapeType of part to mesh)",
                value=["From Shape", "1D", "2D", "3D"],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="ElementOrder",
                group="Mesh Parameters",
                doc="Order of mesh elements",
                value=["1st", "2nd"],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="OptimizeStd",
                group="Mesh Parameters",
                doc="Optimize tetrahedral elements",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="OptimizeNetgen",
                group="Mesh Parameters",
                doc="Optimize tetra elements by use of Netgen",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="HighOrderOptimize",
                group="Mesh Parameters",
                doc="Optimization of high order meshes",
                value=[
                    "None",
                    "Optimization",
                    "Elastic+Optimization",
                    "Elastic",
                    "Fast curving",
                ],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="RecombineAll",
                group="Mesh Parameters",
                doc="Apply recombination algorithm to all surfaces",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="Recombine3DAll",
                group="Mesh Parameters",
                doc="Apply recombination algorithm to all volumes",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="RecombinationAlgorithm",
                group="Mesh Parameters",
                doc="Recombination algorithm",
                value=[
                    "Simple",
                    "Blossom",
                    "Simple full-quad",
                    "Blossom full-quad",
                ],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="CoherenceMesh",
                group="Mesh Parameters",
                doc="Removes all duplicate mesh vertices",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyFloat",
                name="GeometryTolerance",
                group="Mesh Parameters",
                doc="Geometrical Tolerance (0.0 means GMSH std = 1e-08)",
                value=1e-06,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="SecondOrderLinear",
                group="Mesh Parameters",
                doc="Second order nodes are created by linear interpolation",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyIntegerConstraint",
                name="MeshSizeFromCurvature",
                group="Mesh Parameters",
                doc="Number of elements per 2*pi radians, 0 to deactivate",
                value=(12, 0, 10000, 1),
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="Algorithm2D",
                group="Mesh Parameters",
                doc="Mesh algorithm 2D",
                value=[
                    "Automatic",
                    "MeshAdapt",
                    "Delaunay",
                    "Frontal",
                    "BAMG",
                    "DelQuad",
                    "Packing Parallelograms",
                    "Quasi-structured Quad",
                ],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="Algorithm3D",
                group="Mesh Parameters",
                doc="Mesh algorithm 3D",
                value=[
                    "Automatic",
                    "Delaunay",
                    "New Delaunay",
                    "Frontal",
                    "MMG3D",
                    "R-tree",
                    "HXT",
                ],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="GroupsOfNodes",
                group="Mesh Parameters",
                doc="For each group create not only the elements but the nodes too",
                value=True,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="SubdivisionAlgorithm",
                group="Mesh Parameters",
                doc="Mesh subdivision algorithm",
                value=["None", "All Quadrangles", "All Hexahedra", "Barycentric"],
            )
        )

        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except Base.PropertyError:
                prop.add_to_object(obj)

            if prop.name == "Algorithm2D":
                # refresh the list of known 2D algorithms for old projects
                obj.Algorithm2D = prop.value
            elif prop.name == "Algorithm3D":
                # refresh the list of known 3D algorithms for old projects
                obj.Algorithm3D = prop.value
            elif prop.name == "HighOrderOptimize":
                # HighOrderOptimize was once App::PropertyBool, so check this
                prop.handle_change_type(
                    obj, "App::PropertyBool", lambda x: "Optimization" if x else "None"
                )
            # Migrate group of properties for old projects
            if obj.getGroupOfProperty(prop.name) == "FEM Gmsh Mesh Params":
                obj.setGroupOfProperty(prop.name, "Mesh Parameters")

        # migrate old Part property to Shape property
        try:
            value_part = obj.getPropertyByName("Part")
            obj.setPropertyStatus("Part", "-LockDynamic")
            obj.removeProperty("Part")
            # old object is Fem::FemMeshObjectPython (does not have Shape property with global scope)
            prop = _PropHelper(
                type="App::PropertyLinkGlobal",
                name="Shape",
                group="FEM Mesh",
                doc="Geometry object, the mesh is made from. The geometry object has to have a Shape.",
                value=value_part,
            )
            prop.add_to_object(obj)
        except Base.PropertyError:
            pass



# ***************************************************************************
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint initial pressure task panel for the document object"
__author__ = "Uwe Stöhr"
__url__ = "https://www.freecad.org"

## @package task_constraint_initialpressure
#  \ingroup FEM
#  \brief task panel for constraint initial pressure object

import FreeCAD
import FreeCADGui
from femguiutils import selection_widgets

from femtools import membertools
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):

    def __init__(self, obj):
        super().__init__(obj)

        self._paramWidget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/InitialPressure.ui"
        )
        self._initParamWidget()

        # geometry selection widget
        # start with Solid in list!
        self._selectionWidget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face"], True, False
        )

        # form made from param and selection widget
        self.form = [self._paramWidget, self._selectionWidget]

        analysis = obj.getParentGroup()
        self._mesh = None
        self._part = None
        if analysis is not None:
            self._mesh = membertools.get_single_member(analysis, "Fem::FemMeshObject")
        if self._mesh is not None:
            self._part = self._mesh.Shape
        self._partVisible = None
        self._meshVisible = None

    def open(self):
        if self._mesh is not None and self._part is not None:
            self._meshVisible = self._mesh.ViewObject.isVisible()
            self._partVisible = self._part.ViewObject.isVisible()
            self._mesh.ViewObject.hide()
            self._part.ViewObject.show()

    def reject(self):
        self._restoreVisibility()
        self._selectionWidget.finish_selection()
        return super().reject()

    def accept(self):
        if self.obj.References != self._selectionWidget.references:
            self.obj.References = self._selectionWidget.references
        self._applyWidgetChanges()
        self._selectionWidget.finish_selection()
        self._restoreVisibility()
        return super().accept()

    def _restoreVisibility(self):
        if self._mesh is not None and self._part is not None:
            if self._meshVisible:
                self._mesh.ViewObject.show()
            else:
                self._mesh.ViewObject.hide()
            if self._partVisible:
                self._part.ViewObject.show()
            else:
                self._part.ViewObject.hide()

    def _initParamWidget(self):
        self._paramWidget.pressureQSB.setProperty("value", self.obj.Pressure)
        FreeCADGui.ExpressionBinding(self._paramWidget.pressureQSB).bind(self.obj, "Pressure")

    def _applyWidgetChanges(self):
        pressure = None
        try:
            pressure = self._paramWidget.pressureQSB.property("value")
        except ValueError:
            FreeCAD.Console.PrintMessage(
                "Wrong input. Not recognised input: '{}' "
                "Pressure has not been set.\n".format(self._paramWidget.pressureQSB.text())
            )
        if pressure is not None:
            self.obj.Pressure = pressure



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Ellipse object."""
## @package ellipse
# \ingroup draftobjects
# \brief Provides the object code for the Ellipse object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftobjects.base import DraftObject
from draftutils import gui_utils
from draftutils import params


class Ellipse(DraftObject):
    """The Circle object"""

    def __init__(self, obj):
        super().__init__(obj, "Ellipse")

        _tip = QT_TRANSLATE_NOOP("App::Property", "Start angle of the elliptical arc")
        obj.addProperty("App::PropertyAngle", "FirstAngle", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP(
            "App::Property",
            "End angle of the elliptical arc \n\
                (for a full circle, give it same value as First Angle)",
        )
        obj.addProperty("App::PropertyAngle", "LastAngle", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Minor radius of the ellipse")
        obj.addProperty("App::PropertyLength", "MinorRadius", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Major radius of the ellipse")
        obj.addProperty("App::PropertyLength", "MajorRadius", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Create a face")
        obj.addProperty("App::PropertyBool", "MakeFace", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Area of this object")
        obj.addProperty("App::PropertyArea", "Area", "Draft", _tip, locked=True)

        obj.MakeFace = params.get_param("MakeFaceMode")

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(obj, vp_module="view_base", vp_class="ViewProviderDraft")

    def execute(self, obj):
        if self.props_changed_placement_only():
            obj.positionBySupport()
            self.props_changed_clear()
            return

        import Part

        plm = obj.Placement
        if obj.MajorRadius.Value < obj.MinorRadius.Value:
            _err = "Error: Major radius is smaller than the minor radius"
            App.Console.PrintMessage(QT_TRANSLATE_NOOP("Draft", _err))
            return
        if obj.MajorRadius.Value and obj.MinorRadius.Value:
            ell = Part.Ellipse(App.Vector(0, 0, 0), obj.MajorRadius.Value, obj.MinorRadius.Value)
            shape = ell.toShape()
            if hasattr(obj, "FirstAngle"):
                if obj.FirstAngle.Value != obj.LastAngle.Value:
                    a1 = obj.FirstAngle.getValueAs(App.Units.Radian)
                    a2 = obj.LastAngle.getValueAs(App.Units.Radian)
                    shape = Part.ArcOfEllipse(ell, a1, a2).toShape()
            shape = Part.Wire(shape)
            if shape.isClosed():
                if hasattr(obj, "MakeFace"):
                    if obj.MakeFace:
                        shape = Part.Face(shape)
                else:
                    shape = Part.Face(shape)
            obj.Shape = shape
            if hasattr(obj, "Area") and hasattr(shape, "Area"):
                obj.Area = shape.Area
            obj.Placement = plm
        obj.positionBySupport()
        self.props_changed_clear()

    def onChanged(self, obj, prop):
        self.props_changed_store(prop)


# Alias for compatibility with v0.18 and earlier
_Ellipse = Ellipse

## @}



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Mesh export for Nastran mesh file format"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package exportPyNastranMesh
#  \ingroup FEM
#  \brief FreeCAD pyNastran Mesh writer for FEM workbench

import FreeCAD

# we need to import FreeCAD before the non FreeCAD library because of the print
try:
    from pyNastran.bdf.bdf import BDF
except Exception:
    FreeCAD.Console.PrintError(
        "Module pyNastran not found. Writing Mystran solver input will not work.\n"
    )

from FreeCAD import Console

from femmesh import meshtools


# ************************************************************************************************
# ********* generic FreeCAD export methods *******************************************************
# names are fix given from FreeCAD, these methods are called from FreeCAD
# they are set in FEM modules Init.py


def export(objectslist, filename):
    "called when freecad exports a file"
    if len(objectslist) != 1:
        Console.PrintError("This exporter can only export one object.\n")
        return
    obj = objectslist[0]
    if not obj.isDerivedFrom("Fem::FemMeshObject"):
        Console.PrintError("No FEM mesh object selected.\n")
        return
    write(obj.FemMesh, filename)


# ************************************************************************************************
# ********* module specific methods **************************************************************
# writer:
# - a method directly writes a FemMesh to the mesh file
# - a method generates the pyNastran code


# ********* writer *******************************************************************************
def write(fem_mesh, filename):
    """directly write a FemMesh to a pyNastran mesh file format
    fem_mesh: a FemMesh"""

    if not fem_mesh.isDerivedFrom("Fem::FemMesh"):
        Console.PrintError("Not a FemMesh was given as parameter.\n")
        return
    femnodes_mesh = fem_mesh.Nodes
    femelement_table = meshtools.get_femelement_table(fem_mesh)
    export_element_type = get_export_element_type(fem_mesh, femelement_table)

    model = BDF()
    mesh_pynas_code = get_pynastran_mesh(femnodes_mesh, femelement_table, export_element_type)
    mesh_pynas_code += missing_code_pnynasmesh

    # pynas file
    basefilename = filename[: len(filename) - 4]  # TODO basename is more failsafe
    pynasf = open(basefilename + ".py", "w")
    pynasf.write("# written by FreeCAD\n\n\n")
    pynasf.write("from pyNastran.bdf.bdf import BDF\n")
    pynasf.write("model = BDF()\n\n\n")
    pynasf.write(mesh_pynas_code)

    pynasf.write("model.write_bdf('{}', enddata=True)\n".format(basefilename + "_pyNas.bdf"))
    pynasf.close()

    # execute pyNastran code to add grid to the model
    # print(model)
    # print(model.get_bdf_stats())
    exec(mesh_pynas_code)
    # print(model)
    # print(model.get_bdf_stats())

    # write Nastran mesh file
    model.write_bdf(filename, enddata=True)  # TODO FIXME "BEGIN BULK" is missing


def get_pynastran_mesh(
    femnodes_mesh,
    femelement_table,
    export_element_type,
):
    if export_element_type is None:
        Console.PrintError("Error: wrong export_element_type.\n")
        return

    # nodes
    pynas_nodes = "# grid cards, geometric mesh points\n"
    for node in femnodes_mesh:
        vec = femnodes_mesh[node]
        pynas_nodes += f"model.add_grid({node}, [{vec.x}, {vec.y}, {vec.z}])\n"
    # print(pynas_nodes)

    # elements
    # Nastran seems to have the same node order as SMESH (FreeCAD) has
    # thus just write the nodes at once
    pynas_elements = "# elements cards\n"
    for element in femelement_table:
        nodes = femelement_table[element]
        # print(element)  #  eleid
        # print(n)  # tuple of nodes
        if export_element_type == "cbar":
            pynas_elements += (
                "model.add_{ele_keyword}({eid}, {pid}, {nodes}, "
                "{orientation_vec}, {gnull})\n".format(
                    ele_keyword=export_element_type,
                    eid=element,
                    pid=1,
                    nodes=nodes,
                    orientation_vec="x=[0.0, 0.0, 1.0]",
                    gnull="g0=None",
                )
            )
        else:
            if export_element_type == "ctetra4":
                ele_keyword = "ctetra"
                # N1, N3, N2, N4
                the_nodes = [nodes[0], nodes[2], nodes[1], nodes[3]]
            elif export_element_type == "ctetra10":
                ele_keyword = "ctetra"
                # N1, N3, N2, N4, N7, N6, N5, N8, N10, N9
                the_nodes = [
                    nodes[0],
                    nodes[2],
                    nodes[1],
                    nodes[3],
                    nodes[6],
                    nodes[5],
                    nodes[4],
                    nodes[7],
                    nodes[9],
                    nodes[8],
                ]
            else:
                ele_keyword = export_element_type
                the_nodes = nodes
            pynas_elements += "model.add_{ele_keyword}({eid}, {pid}, {nodes})\n".format(
                ele_keyword=ele_keyword, eid=element, pid=1, nodes=the_nodes
            )
    # print(pynas_elements)

    mesh_pynas_code = f"{pynas_nodes}\n\n{pynas_elements}\n\n"
    return mesh_pynas_code


# Helper
def get_export_element_type(femmesh, femelement_table=None):
    return nastran_ele_types[meshtools.get_femmesh_eletype(femmesh, femelement_table)]


nastran_ele_types = {
    "tetra4": "ctetra4",
    "tetra10": "ctetra10",
    "hexa8": None,
    "hexa20": None,
    "tria3": "ctria3",
    "tria6": None,
    "quad4": "cquad4",
    "quad8": None,
    "seg2": "cbar",
    "seg3": None,
    "None": None,
}


missing_code_pnynasmesh = """
model.sol = 101  # is this needed?

# case control
from pyNastran.bdf.bdf import CaseControlDeck
cc = CaseControlDeck([
    #"ECHO = NONE",
    "TITLE = pyNastran for generating solverinput for for Mystran",
    #"SUBCASE 1",
])
model.case_control_deck = cc\n\n
"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the Text object."""
## @package view_text
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the Text object.

## \addtogroup draftviewproviders
# @{
import pivy.coin as coin
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
from draftutils import gui_utils
from draftutils import params
from draftutils.translate import translate
from draftviewproviders.view_draft_annotation import ViewProviderDraftAnnotation


class ViewProviderText(ViewProviderDraftAnnotation):
    """Viewprovider for the Draft Text annotation."""

    def set_text_properties(self, vobj, properties):
        """Set text properties only if they don't already exist."""
        super().set_text_properties(vobj, properties)

        if "Justification" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Horizontal alignment")
            vobj.addProperty("App::PropertyEnumeration", "Justification", "Text", _tip, locked=True)
            vobj.Justification = ["Left", "Center", "Right"]

        if "LineSpacing" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Line spacing (relative to font size)")
            vobj.addProperty("App::PropertyFloat", "LineSpacing", "Text", _tip, locked=True)
            vobj.LineSpacing = params.get_param("LineSpacing")

    def getIcon(self):
        """Return the path to the icon used by the view provider."""
        return ":/icons/Draft_Text.svg"

    def attach(self, vobj):
        """Set up the scene sub-graph of the view provider."""
        self.Object = vobj.Object

        # Main attributes of the Coin scenegraph
        self.mattext = coin.SoMaterial()
        self.font = coin.SoFont()
        self.text_wld = coin.SoAsciiText()  # World orientation. Can be oriented in 3D space.
        self.text_scr = coin.SoText2()  # Screen orientation. Always faces the camera.

        textdrawstyle = coin.SoDrawStyle()
        textdrawstyle.style = coin.SoDrawStyle.FILLED

        # The text string needs to be initialized to something,
        # otherwise it may crash
        self.text_wld.string = self.text_scr.string = "Label"
        self.text_wld.justification = coin.SoAsciiText.LEFT
        self.text_scr.justification = coin.SoText2.LEFT

        self.node_wld = coin.SoGroup()
        self.node_wld.addChild(self.mattext)
        self.node_wld.addChild(textdrawstyle)
        self.node_wld.addChild(self.font)
        self.node_wld.addChild(self.text_wld)

        self.node_scr = coin.SoGroup()
        self.node_scr.addChild(self.mattext)
        self.node_scr.addChild(textdrawstyle)
        self.node_scr.addChild(self.font)
        self.node_scr.addChild(self.text_scr)

        vobj.addDisplayMode(self.node_wld, "World")
        vobj.addDisplayMode(self.node_scr, "Screen")
        self.onChanged(vobj, "TextColor")
        self.onChanged(vobj, "FontSize")
        self.onChanged(vobj, "FontName")
        self.onChanged(vobj, "Justification")
        self.onChanged(vobj, "LineSpacing")
        self.onChanged(vobj, "ScaleMultiplier")

    def updateData(self, obj, prop):
        """Execute when a property from the Proxy class is changed."""
        if prop == "Text" and obj.Text:
            self.text_wld.string.setValue("")
            self.text_scr.string.setValue("")
            _list = [l for l in obj.Text if l]
            self.text_wld.string.setValues(_list)
            self.text_scr.string.setValues(_list)

        elif prop == "Placement":
            transform = gui_utils.find_coin_node(obj.ViewObject.RootNode, coin.SoTransform)
            if transform:
                transform.translation.setValue(obj.Placement.Base)
                transform.rotation.setValue(obj.Placement.Rotation.Q)

    def onChanged(self, vobj, prop):
        """Execute when a view property is changed."""
        super().onChanged(vobj, prop)

        properties = vobj.PropertiesList

        if prop == "ScaleMultiplier" and "ScaleMultiplier" in properties and vobj.ScaleMultiplier:
            if "FontSize" in properties:
                self.font.size = vobj.FontSize.Value * vobj.ScaleMultiplier

        elif prop == "TextColor" and "TextColor" in properties:
            col = vobj.TextColor
            self.mattext.diffuseColor.setValue([col[0], col[1], col[2]])

        elif prop == "FontName" and "FontName" in properties:
            self.font.name = vobj.FontName.encode("utf8")

        elif (
            prop == "FontSize"
            and "FontSize" in properties
            and "ScaleMultiplier" in properties
            and vobj.ScaleMultiplier
        ):
            # In v0.19 this code causes an `AttributeError` exception
            # during loading of the document as `ScaleMultiplier`
            # apparently isn't set immediately when the document loads.
            # So the if condition tests for the existence
            # of `ScaleMultiplier` before even changing the font size.
            # A try-except block may be used as well to catch and ignore
            # this error.
            self.font.size = vobj.FontSize.Value * vobj.ScaleMultiplier

        elif prop == "Justification" and "Justification" in properties:
            # This code was surrounded by a try-except block in order to catch
            # and ignore an `AssertionError` exception. This was the result
            # of a race condition that would result in the `Justification`
            # property not being set at the time that this code is invoked.
            # However, in v0.19 this problem doesn't seem to exist any more
            # so we removed the try-except block.
            if vobj.Justification == "Left":
                self.text_wld.justification = coin.SoAsciiText.LEFT
                self.text_scr.justification = coin.SoText2.LEFT
            elif vobj.Justification == "Right":
                self.text_wld.justification = coin.SoAsciiText.RIGHT
                self.text_scr.justification = coin.SoText2.RIGHT
            else:
                self.text_wld.justification = coin.SoAsciiText.CENTER
                self.text_scr.justification = coin.SoText2.CENTER

        elif prop == "LineSpacing" and "LineSpacing" in properties:
            self.text_wld.spacing = vobj.LineSpacing
            self.text_scr.spacing = vobj.LineSpacing

    def edit(self):
        if not hasattr(Gui, "draftToolBar"):
            import DraftGui
        self.text = ""
        Gui.draftToolBar.sourceCmd = self
        Gui.draftToolBar.taskUi(title=translate("draft", "Text"), icon="Draft_Text")
        Gui.draftToolBar.textUi()
        Gui.draftToolBar.continueCmd.hide()
        Gui.draftToolBar.textValue.setPlainText("\n".join(self.Object.Text))
        Gui.draftToolBar.textValue.setFocus()

    def createObject(self):
        if hasattr(self, "Object"):
            txt = self.text
            if not txt:
                self.finish()
                return None
            # If the last element is an empty string "" we remove it
            if not txt[-1]:
                txt.pop()
            t_list = ['"' + l + '"' for l in txt]
            list_as_text = ", ".join(t_list)
            string = "[" + list_as_text + "]"
            Gui.doCommand("FreeCAD.ActiveDocument." + self.Object.Name + ".Text = " + string)
            App.ActiveDocument.recompute()
            self.finish()

    def finish(self, cont=False):
        Gui.draftToolBar.sourceCmd = None
        Gui.draftToolBar.offUi()


# Alias for compatibility with v0.18 and earlier
ViewProviderDraftText = ViewProviderText

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Polygon object."""
## @package polygon
# \ingroup draftobjects
# \brief Provides the object code for the Polygon object.

## \addtogroup draftobjects
# @{
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import DraftGeomUtils
from draftobjects.base import DraftObject
from draftutils import gui_utils
from draftutils import params


class Polygon(DraftObject):
    """The Polygon object"""

    def __init__(self, obj):
        super().__init__(obj, "Polygon")

        _tip = QT_TRANSLATE_NOOP("App::Property", "Number of faces")
        obj.addProperty("App::PropertyInteger", "FacesNumber", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Radius of the control circle")
        obj.addProperty("App::PropertyLength", "Radius", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP(
            "App::Property", "How the polygon must be drawn from the control circle"
        )
        obj.addProperty("App::PropertyEnumeration", "DrawMode", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Radius to use to fillet the corners")
        obj.addProperty("App::PropertyLength", "FilletRadius", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Size of the chamfer to give to the corners")
        obj.addProperty("App::PropertyLength", "ChamferSize", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Create a face")
        obj.addProperty("App::PropertyBool", "MakeFace", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The area of this object")
        obj.addProperty("App::PropertyArea", "Area", "Draft", _tip, locked=True)

        obj.MakeFace = params.get_param("MakeFaceMode")
        obj.DrawMode = ["inscribed", "circumscribed"]
        obj.FacesNumber = 0
        obj.Radius = 1

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(obj, vp_module="view_base", vp_class="ViewProviderDraft")

    def execute(self, obj):
        if self.props_changed_placement_only():
            obj.positionBySupport()
            self.props_changed_clear()
            return

        if (obj.FacesNumber >= 3) and (obj.Radius.Value > 0):
            import Part

            plm = obj.Placement
            angle = (math.pi * 2) / obj.FacesNumber
            if obj.DrawMode == "inscribed":
                delta = obj.Radius.Value
            else:
                delta = obj.Radius.Value / math.cos(angle / 2.0)
            pts = [App.Vector(delta, 0, 0)]
            for i in range(obj.FacesNumber - 1):
                ang = (i + 1) * angle
                pts.append(App.Vector(delta * math.cos(ang), delta * math.sin(ang), 0))
            pts.append(pts[0])
            shape = Part.makePolygon(pts)
            if "ChamferSize" in obj.PropertiesList:
                if obj.ChamferSize.Value != 0:
                    w = DraftGeomUtils.filletWire(shape, obj.ChamferSize.Value, chamfer=True)
                    if w:
                        shape = w
            if "FilletRadius" in obj.PropertiesList:
                if obj.FilletRadius.Value != 0:
                    w = DraftGeomUtils.filletWire(shape, obj.FilletRadius.Value)
                    if w:
                        shape = w
            if hasattr(obj, "MakeFace"):
                if obj.MakeFace:
                    shape = Part.Face(shape)
            else:
                shape = Part.Face(shape)
            obj.Shape = shape
            if hasattr(obj, "Area") and hasattr(shape, "Area"):
                obj.Area = shape.Area
            obj.Placement = plm
        obj.positionBySupport()
        self.props_changed_clear()

    def onChanged(self, obj, prop):
        self.props_changed_store(prop)


# Alias for compatibility with v0.18 and earlier
_Polygon = Polygon

## @}
